"""Program schedule."""
from datetime import datetime
from typing import Optional

from radikopodcast.database.database import Database
from radikopodcast.database.models import Program, Station
from radikopodcast.database.program_downloader import ProgramDownloader
from radikopodcast.radiko_datetime import RadikoDatetime
from radikopodcast.radikoapi.radiko_api import RadikoApi
from radikopodcast.radikoxml.xml_converter import XmlConverterStation


class ProgramSchedule:
    """Program schedule."""

    AREA_ID_DEFAULT = "JP13"  # TOKYO
    HOUR_WHEN_RADIKO_PROGRAM_UPDATE = 5
    MINUTE_MARGIN_WHEN_RADIKO_PROGRAM_UPDATE = 15

    def __init__(self, *, area_id: str = AREA_ID_DEFAULT) -> None:
        # pylint bug, see: https://github.com/PyCQA/pylint/issues/3882
        # pylint: disable=unsubscriptable-object
        self.last_updated: Optional[datetime] = None
        self.area_id = area_id
        self.database = Database()
        if Station.is_empty():
            Station.save_all(XmlConverterStation(RadikoApi(area_id=self.area_id).get_station()).to_model())

    @staticmethod
    def search(keywords: list[str]) -> list[Program]:
        return Program.find(keywords)

    def download_if_program_has_not_been_downloaded(self) -> None:
        now = RadikoDatetime.now_jst()
        if self.has_downloaded(now):
            return
        self.add(now)
        self.remove(now)
        self.last_updated = now

    def has_downloaded(self, now: datetime) -> bool:
        return (
            self.last_updated is not None
            and RadikoDatetime.is_same_radiko_date(now, self.last_updated)
            or now.hour == self.HOUR_WHEN_RADIKO_PROGRAM_UPDATE
            and now.minute <= self.MINUTE_MARGIN_WHEN_RADIKO_PROGRAM_UPDATE
        )

    def add(self, now: datetime) -> None:
        """Adds programs from radiko API."""
        program_downloader = ProgramDownloader(now, area_id=self.area_id)
        program_downloader.download_all_time_free_programs()

    @staticmethod
    def remove(now: datetime) -> None:
        Program.delete(RadikoDatetime.time_free_oldest_date(now))



"""Datetime for radiko specification."""
from datetime import date, datetime, timedelta, timezone

JST = timezone(timedelta(hours=+9), "JST")


class RadikoDatetime:
    """Datetime for radiko specification."""

    FORMAT_CODE = "%Y%m%d%H%M%S"

    @staticmethod
    def encode(argument_datetime: datetime) -> str:
        return argument_datetime.strftime(RadikoDatetime.FORMAT_CODE)

    @staticmethod
    def decode(argument_string: str) -> datetime:
        return datetime.strptime(argument_string, RadikoDatetime.FORMAT_CODE).replace(tzinfo=JST)

    @staticmethod
    def time_free_oldest_date(now: datetime) -> date:
        return (now - timedelta(days=7, hours=5)).date()

    @staticmethod
    def time_free_day_before_newest_date(now: datetime) -> date:
        return (now - timedelta(days=1, hours=5)).date()

    @staticmethod
    def now_jst() -> datetime:
        return datetime.now(tz=JST)

    @staticmethod
    def is_same_radiko_date(datetime_a: datetime, datetime_b: datetime) -> bool:
        return (datetime_a - timedelta(hours=5)).date() == (datetime_b - timedelta(hours=5)).date()


class RadikoDate:
    """Date for radiko specification."""

    FORMAT_CODE = "%Y%m%d"

    @staticmethod
    def encode(argument_datetime: date) -> str:
        return argument_datetime.strftime(RadikoDate.FORMAT_CODE)

    @staticmethod
    def decode(argument_string: str) -> date:
        return datetime.strptime(argument_string, RadikoDate.FORMAT_CODE).replace(tzinfo=JST).date()



"""Console script for radikopodcast."""
import asyncio
from pathlib import Path
import sys
from typing import NoReturn, Optional

import click

from radikopodcast.radiko_podcast import RadikoPodcast


@click.command()
@click.option(
    "-f",
    "--file",
    type=click.Path(exists=True, resolve_path=True, path_type=Path),
    default=Path("config.yml"),
)
def radiko_podcast(file: Optional[Path]) -> NoReturn:
    """Console script for radiko Podcast."""
    podcast = RadikoPodcast(path_to_configuration=file)
    try:
        podcast.run()
    except (KeyboardInterrupt, asyncio.CancelledError):
        sys.exit(130)
    sys.exit(1)  # pragma: no cover



"""radiko API."""
from logging import getLogger
from typing import TYPE_CHECKING

from defusedxml import ElementTree

from radikopodcast.radikoapi.requester import Requester

if TYPE_CHECKING:
    from datetime import date

    # Reason: The defusedxml's issue:
    # - defusedxml lacks an Element class · Issue #48 · tiran/defusedxml
    #   https://github.com/tiran/defusedxml/issues/48
    # nosemgrep: python.lang.security.use-defused-xml.use-defused-xml  # noqa: ERA001
    from xml.etree.ElementTree import Element  # nosec B405


class RadikoApi:
    """radiko API."""

    AREA_ID_DEFAULT = "JP13"  # TOKYO

    def __init__(self, *, area_id: str = AREA_ID_DEFAULT) -> None:
        self.area_id = area_id
        self.logger = getLogger(__name__)

    def get_program(self, target_date: "date") -> "Element":
        return self.get(f"https://radiko.jp/v3/program/date/{target_date.strftime('%Y%m%d')}/{self.area_id}.xml")

    def get_station(self) -> "Element":
        return self.get(f"https://radiko.jp/v3/station/list/{self.area_id}.xml")

    @staticmethod
    def get(url: str) -> "Element":
        # Reason: The defusedxml's responsible.
        return ElementTree.fromstring(Requester.get(url).content, forbid_dtd=True)  # type: ignore[no-any-return]



"""XML converter."""
from abc import abstractmethod
from logging import getLogger
from typing import Generic, Optional, TYPE_CHECKING

from radikopodcast.database.models import Program, Station
from radikopodcast.database.models import TypeVarModelInitByXml as T  # noqa: N814  Reason: TypeVar.
from radikopodcast.radikoxml.xml_parser import XmlParser, XmlParserProgram, XmlParserStation

if TYPE_CHECKING:
    from datetime import date

    # Reason: The defusedxml's issue:
    # - defusedxml lacks an Element class · Issue #48 · tiran/defusedxml
    #   https://github.com/tiran/defusedxml/issues/48
    # nosemgrep: python.lang.security.use-defused-xml.use-defused-xml  # noqa: ERA001
    from xml.etree.ElementTree import Element  # nosec B405


class XmlConverter(Generic[T]):
    """XML converter."""

    def __init__(self) -> None:
        self.logger = getLogger(__name__)

    def to_model(self) -> list[T]:
        """Converts to list of SQLAlchemy model of station of radiko."""
        list_model = list(filter(lambda x: x is not None, self.comprehension_notation()))
        self.logger.debug(list_model)
        # Reason: mypy's bug, see: https://github.com/python/mypy/issues/6847
        return list_model  # type: ignore[return-value]

    @abstractmethod
    def comprehension_notation(self) -> list[Optional[T]]:
        raise NotImplementedError  # pragma: no cover

    def create_model(self, xml_parser: XmlParser, model_class: type[T]) -> Optional[T]:
        if xml_parser.validate:
            self.logger.error(xml_parser.list_error)
            return None
        return model_class(xml_parser)


class XmlConverterProgram(XmlConverter[Program]):
    """XML converter for program of radiko."""

    def __init__(self, target_date: "date", element_tree_radiko: "Element", area_id: str) -> None:
        super().__init__()
        self.target_date = target_date
        self.element_tree_radiko = element_tree_radiko
        self.area_id = area_id

    def comprehension_notation(self) -> list[Optional[Program]]:
        return [
            self.create_model(self.create_xml_parser(element_tree_program, element_tree_station), Program)
            for element_tree_station in self.element_tree_radiko.findall("./stations/station")
            for element_tree_program in element_tree_station.findall("./progs/prog")
        ]

    def create_xml_parser(self, element_tree_program: "Element", element_tree_station: "Element") -> XmlParserProgram:
        return XmlParserProgram(element_tree_program, element_tree_station, self.target_date, self.area_id)


class XmlConverterStation(XmlConverter[Station]):
    """XML converter for station of radiko."""

    def __init__(self, element_tree_list_station: "Element") -> None:
        super().__init__()
        self.element_tree_list_station = element_tree_list_station

    def comprehension_notation(self) -> list[Optional[Station]]:
        return [
            self.create_model(XmlParserStation(element_tree_station), Station)
            for element_tree_station in self.element_tree_list_station.findall("./station")
        ]



"""Tests for radiko_stream_spec_factory.py."""
from dataclasses import dataclass
from typing import Any, TYPE_CHECKING

# Reason: Maybe, requires to update ffmpeg-python side.
from ffmpeg.nodes import Stream  # type: ignore[import]
import pytest

from radikopodcast.radiko_stream_spec_factory import RadikoStreamSpecFactory

if TYPE_CHECKING:
    from pathlib import Path

    from radikopodcast.database.models import Program


@dataclass
class MasterPlaylist:
    media_playlist_url: str
    headers: dict[str, Any]


class TestRadikoStreamSpecFactory:
    """Tests for RadikoStreamSpecFactory."""

    @pytest.mark.asyncio()
    @pytest.mark.usefixtures("mock_master_playlist_client")
    async def test(self, model_program: "Program") -> None:
        """Function create() should create stream spec."""
        radiko_stream_spec_factory = RadikoStreamSpecFactory(model_program)
        stream_spec = await radiko_stream_spec_factory.create()
        assert isinstance(stream_spec, Stream)
        self.check_stream_spec(stream_spec)

    def check_stream_spec(self, stream_spec: Stream) -> None:
        assert "output(c='copy', filename='./output/20210116050000_FMJ_ZAPPA.m4a', format='mp4')[None]" in str(
            stream_spec,
        )
        assert stream_spec.node.short_repr == "20210116050000_FMJ_ZAPPA.m4a"
        assert stream_spec.label is None
        assert stream_spec.selector is None

    @staticmethod
    def test_area_id_none(model_program_area_id_none: "Program") -> None:
        with pytest.raises(ValueError, match="None") as excinfo:
            RadikoStreamSpecFactory(model_program_area_id_none)
        assert "None" in str(excinfo.value)

    @staticmethod
    @pytest.mark.asyncio()
    @pytest.mark.usefixtures("mock_master_playlist_client")
    async def test_file_exists_error(execution_environment: "Path", model_program: "Program") -> None:
        """Method: create() should raise FileExistsError when output file already exists."""
        (execution_environment / "output" / "20210116050000_FMJ_ZAPPA.m4a").touch()
        radiko_stream_spec_factory = RadikoStreamSpecFactory(model_program)
        with pytest.raises(FileExistsError):
            await radiko_stream_spec_factory.create()



"""Top-level package for radiko Podcast."""

__author__ = """Master"""
__email__ = "roadmasternavi@gmail.com"
__version__ = "1.0.2"

from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker

from radikopodcast.config import Config

# pylint: disable=invalid-name
Session = scoped_session(
    sessionmaker(
        bind=create_engine("sqlite:///programs.db"),
        # ↓ @ see https://stackoverflow.com/questions/32922210/why-does-a-query-invoke-a-auto-flush-in-sqlalchemy
        autoflush=False,
        # ↓ To use with-statement
        autocommit=True,
    ),
)
CONFIG: Config = Config()



"""XML parsers."""
from typing import Any, Callable, TYPE_CHECKING

from defusedxml import ElementTree
from errorcollector import MultipleErrorCollector

from radikopodcast.exceptions import XmlParseError
from radikopodcast.radiko_datetime import RadikoDatetime

if TYPE_CHECKING:
    from datetime import date, datetime

    # Reason: The defusedxml's issue:
    # - defusedxml lacks an Element class · Issue #48 · tiran/defusedxml
    #   https://github.com/tiran/defusedxml/issues/48
    # nosemgrep: python.lang.security.use-defused-xml.use-defused-xml  # noqa: ERA001
    from xml.etree.ElementTree import Element  # nosec B405


class XmlParser:
    """Abstract XML parser."""

    def __init__(self) -> None:
        self.list_error: list[XmlParseError] = []

    @property
    def validate(self) -> bool:
        """This method validates data."""
        return bool(self.list_error)

    def stock_error(self, method: Callable[[], Any], message: str) -> Any:
        """This method stocks error."""
        with MultipleErrorCollector(XmlParseError, message, self.list_error):
            return method()

    @staticmethod
    def to_string(element: "Element") -> str:
        # Reason: The defusedxml's responsible.
        return ElementTree.tostring(element, encoding="unicode")  # type: ignore[no-any-return]


# Reason: This class converts argument of constructor to property. pylint: disable=too-few-public-methods
class XmlParserProgram(XmlParser):
    """XML parser for program of radiko."""

    def __init__(
        self,
        element_tree_program: "Element",
        element_tree_station: "Element",
        target_date: "date",
        area_id: str,
    ) -> None:
        super().__init__()
        self.element_tree_program = element_tree_program
        self.element_tree_station = element_tree_station
        self.date = target_date
        self.area_id = area_id

    @property
    # Reason: "id" meets requirement of snake_case. pylint: disable=invalid-name
    def id(self) -> str:  # noqa: A003
        return self.element_tree_program.attrib["id"]

    @property
    # Reason: Can't understand what "ft" points. pylint: disable=invalid-name
    def ft(self) -> "datetime":
        return RadikoDatetime.decode(self.element_tree_program.attrib["ft"])

    @property
    # Reason: "to" meets requirement of snake_case. pylint: disable=invalid-name
    def to(self) -> "datetime":
        return RadikoDatetime.decode(self.element_tree_program.attrib["to"])

    @property
    def title(self) -> str:
        """Title if find it, otherwise, raises error."""
        element_title = self.element_tree_program.find("title")
        if element_title is None:
            message = f"Can't find title. XML: {self.to_string(self.element_tree_program)}"
            raise XmlParseError(message)
        if element_title.text is None:
            message = f"No title text. XML: {self.to_string(self.element_tree_program)}"
            raise XmlParseError(message)
        return element_title.text

    @property
    def station_id(self) -> str:
        return self.element_tree_station.attrib["id"]

    @property
    def validate(self) -> bool:
        self.stock_error(lambda: self.id, f"Invalid id. XML: {self.to_string(self.element_tree_program)}")
        self.stock_error(lambda: self.ft, f"Invalid ft. XML: {self.to_string(self.element_tree_program)}")
        self.stock_error(lambda: self.to, f"Invalid to. XML: {self.to_string(self.element_tree_program)}")
        self.stock_error(lambda: self.title, f"Invalid title. XML: {self.to_string(self.element_tree_program)}")
        self.stock_error(
            lambda: self.station_id,
            f"Invalid station id. XML: {self.to_string(self.element_tree_station)}",
        )
        return super().validate


# Reason: This class converts argument of constructor to property. pylint: disable=too-few-public-methods
class XmlParserStation(XmlParser):
    """XML parser for station of radiko."""

    def __init__(self, element_tree_station: "Element") -> None:
        super().__init__()
        self.element_tree_station = element_tree_station

    @property
    # Reason: "id" meets requirement of snake_case. pylint: disable=invalid-name
    def id(self) -> str:  # noqa: A003
        """ID if find it, otherwise, raises error."""
        element_id = self.element_tree_station.find("./id")
        if element_id is None:
            message = f"Can't find id. XML: {self.to_string(self.element_tree_station)}"
            raise XmlParseError(message)
        if element_id.text is None:
            message = f"No id text. XML: {self.to_string(self.element_tree_station)}"
            raise XmlParseError(message)
        return element_id.text

    @property
    def name(self) -> str:
        """Name if find it, otherwise, raises error."""
        element_name = self.element_tree_station.find("./name")
        if element_name is None:
            message = f"Can't find name. XML: {self.to_string(self.element_tree_station)}"
            raise XmlParseError(message)
        if element_name.text is None:
            message = f"No name text. XML: {self.to_string(self.element_tree_station)}"
            raise XmlParseError(message)
        return element_name.text

    @property
    def validate(self) -> bool:
        self.stock_error(lambda: self.id, f"Invalid id. XML: {self.to_string(self.element_tree_station)}")
        self.stock_error(lambda: self.name, f"Invalid name. XML: {self.to_string(self.element_tree_station)}")
        return super().validate



"""To unify error check and logging process."""
from http import HTTPStatus
from logging import getLogger

from radikoplaylist.exceptions import BadHttpStatusCodeError, HttpRequestTimeoutError
import requests
from requests import Response, Timeout


class Requester:
    """To unify error check and logging process."""

    @staticmethod
    def get(url: str) -> Response:
        """Get request with error check and logging process."""
        logger = getLogger(__name__)
        try:
            res = requests.get(url=url, timeout=20.0)
        except Timeout as error:
            logger.exception("failed in %s.\nRequest Timeout", url)
            raise HttpRequestTimeoutError("failed in " + url + ".") from error
        if res.status_code != HTTPStatus.OK:
            logger.error("failed in %s.", url)
            logger.error("status_code:%s", res.status_code)
            logger.error("content:%s", res.content)
            raise BadHttpStatusCodeError("failed in " + url + ".")
        logger.debug("Request %s succeed.", url)
        return res



"""Database."""
from logging import getLogger
from typing import cast

from sqlalchemy.engine.reflection import Inspector
from sqlalchemy import inspect

from radikopodcast.database.models import Base
from radikopodcast import Session


class Database:
    """Database."""

    def __init__(self) -> None:
        self.logger = getLogger(__name__)
        engine = Session.get_bind()
        inspector = cast(Inspector, inspect(engine))
        if not inspector.has_table("programs"):
            self.initialize_database()

    @staticmethod
    def initialize_database() -> None:
        """This function create empty tables from SQLAlchemy models."""
        # pylint: disable=no-member
        Base.metadata.create_all(Session.get_bind(), checkfirst=False)



"""Radiko Archiver."""
import asyncio
from logging import getLogger
import os

# Reason: Following export method in __init__.py from Effective Python 2nd Edition item 85
from asyncffmpeg import FFmpegCoroutineFactory

from radikopodcast.database.models import Program
from radikopodcast.radiko_stream_spec_factory import RadikoStreamSpecFactory

TIME_TO_FORCE_TERMINATION = 8


class RadikoArchiver:
    """Radiko Archiver."""

    def __init__(
        self,
        *,
        time_to_force_termination: int = TIME_TO_FORCE_TERMINATION,
        stop_if_file_exists: bool = False,
    ) -> None:
        self.ffmpeg_coroutine = FFmpegCoroutineFactory.create(time_to_force_termination=time_to_force_termination)
        self.stop_if_file_exists = stop_if_file_exists

    async def archive(self, program: Program) -> None:
        """Archives radiko program."""
        logger = getLogger(__name__)
        logger.debug("Start archive")
        logger.debug(
            "program time: %s, station: %s, start: %s, end: %s",
            program.title,
            program.station_id,
            program.ft,
            program.to,
        )
        try:
            program.mark_archiving()
            logger.info(self.ffmpeg_coroutine)
            logger.info(self.ffmpeg_coroutine.execute)
            await self.ffmpeg_coroutine.execute(RadikoStreamSpecFactory(program).create)
        except (KeyboardInterrupt, asyncio.CancelledError):
            logger.debug("SIGINT for PID=%d", os.getpid())
            logger.debug("FFmpeg run cancelled.")
            program.mark_suspended()
            raise
        except FileExistsError:
            if self.stop_if_file_exists:
                raise
            program.mark_failed()
            return
        program.mark_archived()



"""This module implements SQLAlchemy session life cycle to prevent forgetting close."""
from contextlib import AbstractContextManager
from types import TracebackType
from typing import Optional

from sqlalchemy.orm.session import Session as SQLAlchemySession

from radikopodcast import Session


class SessionManager(AbstractContextManager[SQLAlchemySession]):
    """This class implements SQLAlchemy session life cycle to prevent forgetting close."""

    def __init__(self) -> None:
        self._session: SQLAlchemySession = Session()

    def __enter__(self) -> SQLAlchemySession:
        return self._session

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self._session.close()



"""Main module."""
import asyncio
import logging
from logging import LogRecord
from pathlib import Path
import queue
from typing import Any, Callable, Optional

# Reason: Following export method in __init__.py from Effective Python 2nd Edition item 85
from asynccpu import ProcessTaskPoolExecutor

from radikopodcast import CONFIG
from radikopodcast.database.program_schedule import ProgramSchedule
from radikopodcast.radiko_archiver import RadikoArchiver, TIME_TO_FORCE_TERMINATION


class RadikoPodcast:
    """Main class."""

    def __init__(
        self,
        *,
        path_to_configuration: Optional[Path] = None,
        time_to_force_termination: int = TIME_TO_FORCE_TERMINATION,
        # Reason: This argument name is API. pylint: disable=redefined-outer-name
        queue: Optional[queue.Queue[LogRecord]] = None,
        configurer: Optional[Callable[[], Any]] = None,
    ) -> None:
        logging.basicConfig(level=logging.DEBUG)
        # Reason Yaml Dataclass Config's bug.
        CONFIG.load(path_to_configuration)  # type: ignore[arg-type]
        self.radiko_archiver = RadikoArchiver(
            time_to_force_termination=time_to_force_termination,
            stop_if_file_exists=CONFIG.stop_if_file_exists,
        )
        self.queue = queue
        self.configurer = configurer
        self.program_schedule = ProgramSchedule(area_id=CONFIG.area_id)
        self.logger = logging.getLogger(__name__)

    def run(self) -> None:
        """Runs."""
        try:
            asyncio.run(self.archive_repeatedly())
        except Exception:
            self.logger.exception("Unexpected error raised!")
            raise

    async def archive_repeatedly(self) -> None:
        """Archives repeatedly."""
        with ProcessTaskPoolExecutor(
            max_workers=CONFIG.number_process,
            cancel_tasks_when_shutdown=True,
            queue=self.queue,
            configurer=self.configurer,
        ) as executor:
            while True:
                self.program_schedule.download_if_program_has_not_been_downloaded()
                for program in self.program_schedule.search(CONFIG.keywords):
                    self.logger.debug("Start: program.title = %s", program.title)
                    # Reason: pylint bug. pylint: disable=no-member
                    executor.create_process_task(self.radiko_archiver.archive, program)
                    self.logger.debug("Finish: program.title = %s", program.title)
                await self.sleep(180)

    @staticmethod
    async def sleep(second: float) -> None:
        await asyncio.sleep(second)



"""Tests for cli.py."""
import re
from unittest import mock

from click.testing import CliRunner
from freezegun.api import freeze_time
import pytest

from radikopodcast import cli
from radikopodcast.radiko_podcast import RadikoPodcast


@pytest.mark.usefixtures("_mock_all")
# Decorator: freeze_time without tick=True breaks EventLoop.run_in_executor().
# see:
#  - Hangs in multiprocessing · Issue #230 · spulec/freezegun
#    https://github.com/spulec/freezegun/issues/230#issuecomment-835721924
@freeze_time("2021-01-17 05:16:00", tz_offset=-9, tick=True)
def test_command_line_interface(runner_in_isolated_filesystem: CliRunner) -> None:
    """Test CLI."""
    with mock.patch.object(RadikoPodcast, "sleep", side_effect=KeyboardInterrupt):
        result = runner_in_isolated_filesystem.invoke(cli.radiko_podcast)
    assert result.output == ""
    # CTRL + C, 128 + SIGINT
    exit_code_ctrl_plus_c = 130
    assert result.exit_code == exit_code_ctrl_plus_c
    help_result = runner_in_isolated_filesystem.invoke(cli.radiko_podcast, ["--help"])
    assert help_result.exit_code == 0
    assert "--help           Show this message and exit." in help_result.output


@pytest.mark.usefixtures("database_session", "_mock_requests_station")
@freeze_time("2021-01-17 05:15:00", tz_offset=-9)
def test_command_line_interface_error(
    runner_in_isolated_filesystem: CliRunner,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test CLI when error."""
    with mock.patch.object(RadikoPodcast, "sleep", side_effect=IOError):
        result = runner_in_isolated_filesystem.invoke(cli.radiko_podcast)
    assert result.output == ""
    assert result.exit_code == 1
    pattern = re.compile(r"ERROR.*OSError$", flags=re.DOTALL | re.MULTILINE)
    assert pattern.search(caplog.text)



"""Test for radiko_archiver.py."""
import asyncio
import sys
from unittest.mock import AsyncMock

from asynccpu.process_task_pool_executor import ProcessTaskPoolExecutor
from asyncffmpeg import FFmpegCoroutine, FFmpegCoroutineFactory
import pytest

from radikopodcast.database.models import Program
from radikopodcast.radiko_archiver import RadikoArchiver


class TestRadikoArchiver:
    """Test for RadikoArchiver."""

    @pytest.mark.usefixtures("record_program", "mock_master_playlist_client", "mock_ffmpeg_coroutine")
    def test(self) -> None:
        self.mock_ffmpeg_coroutine_again()
        program = Program.find(["ROPPONGI PASSION PIT"])[0]
        asyncio.run(RadikoArchiver().archive(program))

    @staticmethod
    @pytest.mark.asyncio()
    @pytest.mark.skipif(sys.platform == "win32", reason="test for Linux only")
    @pytest.mark.usefixtures("record_program", "mock_master_playlist_client", "mock_ffmpeg_coroutine")
    async def test_with_process_task_pool_executor() -> None:
        """ProcessTaskPoolExecutor should shutdown when raise KeyboardInterrupt.

        During running RadikoArchiver.archive().
        """
        program = Program.find(["ROPPONGI PASSION PIT"])[0]
        radiko_archiver = RadikoArchiver()
        # Reason: To test that process can teardown when running asynchronous task
        with pytest.raises(KeyboardInterrupt):  # noqa: PT012, SIM117
            with ProcessTaskPoolExecutor() as executor:
                executor.create_process_task(radiko_archiver.archive, program)
                raise KeyboardInterrupt

    @pytest.mark.skipif(sys.platform == "win32", reason="test for Linux only")
    @pytest.mark.usefixtures("record_program", "mock_master_playlist_client", "mock_ffmpeg_coroutine")
    @pytest.mark.parametrize("type_error", [FileExistsError, KeyboardInterrupt])
    def test_error(self, type_error: type[BaseException]) -> None:
        """Specific error should not catch in method: archive()."""
        ffmpeg_coroutine = self.mock_ffmpeg_coroutine_again()
        # Reason: Creating mock.
        ffmpeg_coroutine.execute.side_effect = type_error  # type: ignore[attr-defined]
        program = Program.find(["ROPPONGI PASSION PIT"])[0]
        with pytest.raises(type_error):
            asyncio.run(RadikoArchiver(stop_if_file_exists=True).archive(program))

    @pytest.mark.usefixtures("record_program", "mock_master_playlist_client", "mock_ffmpeg_coroutine")
    def test_file_exists_skip(self) -> None:
        ffmpeg_coroutine = self.mock_ffmpeg_coroutine_again()
        # Reason: Creating mock.
        ffmpeg_coroutine.execute.side_effect = FileExistsError  # type: ignore[attr-defined]
        program = Program.find(["ROPPONGI PASSION PIT"])[0]
        asyncio.run(RadikoArchiver().archive(program))

    @staticmethod
    def mock_ffmpeg_coroutine_again() -> FFmpegCoroutine:
        """To fix execute as AsyncMock.

        Method execute seems to reset as PicklableMock...
        """
        ffmpeg_coroutine = FFmpegCoroutineFactory.create()
        # Reason: Creating mock.
        ffmpeg_coroutine.execute = AsyncMock()  # type: ignore[method-assign]
        return ffmpeg_coroutine



"""Tests for radiko_podcast.py."""
import asyncio
from pathlib import Path
import sys
from unittest import mock

from freezegun.api import freeze_time
import pytest

from radikopodcast.radiko_podcast import RadikoPodcast


class TestRadikoPodcast:
    """Test for RadikoPodcast."""

    @staticmethod
    @pytest.mark.skipif(sys.platform == "win32", reason="test for Linux only")
    @pytest.mark.usefixtures("_mock_all")
    # Decorator: freeze_time without tick=True breaks EventLoop.run_in_executor().
    # This comment prevents Pylint duplicate-code.
    # see:
    #  - Hangs in multiprocessing · Issue #230 · spulec/freezegun
    #    https://github.com/spulec/freezegun/issues/230#issuecomment-835721924
    @freeze_time("2021-01-17 05:16:00", tz_offset=-9, tick=True)
    def test(config_yaml: Path) -> None:
        """RadikoPodcast should stop when raise KeyboardInterrupt."""
        podcast = RadikoPodcast(path_to_configuration=config_yaml)
        with (
            pytest.raises(KeyboardInterrupt),
            mock.patch.object(RadikoPodcast, "sleep", side_effect=KeyboardInterrupt),
        ):
            podcast.run()

    @staticmethod
    def test_sleep() -> None:
        asyncio.run(RadikoPodcast.sleep(0.1))



"""This module implements SQLAlchemy database models."""
from __future__ import annotations

from abc import abstractmethod

# Reason: For type hint. pylint: disable=unused-import
from enum import IntEnum
from typing import cast, Generic, TYPE_CHECKING, TypeVar

from inflector import Inflector
from sqlalchemy import and_, Column, DATETIME, func, or_, String
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.orm import Mapped, relationship
from sqlalchemy.sql.schema import ForeignKey, MetaData
from sqlalchemy.sql.sqltypes import DATE, INTEGER

from radikopodcast.database.session_manager import SessionManager
from radikopodcast.radiko_datetime import RadikoDatetime
from radikopodcast.radikoxml.xml_parser import XmlParser, XmlParserProgram, XmlParserStation

if TYPE_CHECKING:
    from collections.abc import Iterable
    import datetime

    from sqlalchemy.orm.interfaces import MapperProperty


class ArchiveStatusId(IntEnum):
    """This class implements file for CSV convert id on database."""

    ARCHIVABLE = 0
    ARCHIVING = 1
    SUSPENDED = 2
    FAILED = 3
    ARCHIVED = 4


Base = declarative_base(metadata=MetaData())

TypeVarXmlParser = TypeVar("TypeVarXmlParser", bound=XmlParser)


class ModelInitByXml(Base, Generic[TypeVarXmlParser]):
    """Abstract class of Model which initialized by XML."""

    __abstract__ = True

    @declared_attr
    # Reason: @declared_attr marks method as class level.
    # see: https://docs.sqlalchemy.org/en/14/orm/mapping_api.html#class-mapping-api
    # pylint: disable=no-self-argument
    def __tablename__(cls) -> MapperProperty[str]:  # noqa: N805
        # Reason: "cls" is not instance.
        # pylint: disable=no-member
        return Inflector().pluralize(cls.__name__.lower())

    def __init__(self, xml_parser: TypeVarXmlParser) -> None:
        super().__init__()
        self.init(xml_parser)

    @abstractmethod
    def init(self, xml_parser: TypeVarXmlParser) -> None:
        raise NotImplementedError  # pragma: no cover

    @classmethod
    def save_all(cls, models: Iterable[TypeVarModelInitByXml]) -> None:
        """This method insert Store models into database."""
        with SessionManager() as session, session.begin():
            session.add_all(models)


# Reason: Mypy won't support TypeVar bounding generic type.
# see:
#  - `TypeVar`s cannot refer to type variables · Issue #2756 · python/mypy
#    https://github.com/python/mypy/issues/2756#issuecomment-579496164
TypeVarModelInitByXml = TypeVar("TypeVarModelInitByXml", bound=ModelInitByXml)  # type: ignore[type-arg]


class Station(ModelInitByXml[XmlParserStation]):
    """Station of radiko."""

    # Reason: Design of table column.
    id = Column(String(255), primary_key=True)  # noqa: A003
    name = Column(String(255))
    list_program: Mapped[list[Program]] = relationship("Program", backref="station")
    transfer_target = Column(String(255))

    def init(self, xml_parser: XmlParserStation) -> None:
        # Reason: "id" meets requirement of snake_case. pylint: disable=invalid-name
        self.id = xml_parser.id
        self.name = xml_parser.name

    @staticmethod
    def is_empty() -> bool:
        with SessionManager() as session:
            count = cast(tuple[int], session.query(func.count(Station.id)).one())
            return count == (0,)


class Program(ModelInitByXml[XmlParserProgram]):
    """Program of radiko."""

    # Reason: Model. pylint: disable=too-many-instance-attributes
    id = Column(String(255), primary_key=True)  # noqa: A003
    to = Column(DATETIME)
    ft = Column(DATETIME)
    title = Column(String(255))
    station_id: str = Column(String, ForeignKey("stations.id"), nullable=False)
    date = Column(DATE)
    area_id = Column(String(255))
    archive_status = Column(INTEGER)

    def init(self, xml_parser: XmlParserProgram) -> None:
        # Reason: "id" meets requirement of snake_case. pylint: disable=invalid-name
        self.id = xml_parser.id
        # Reason: Can't understand what "ft" points. pylint: disable=invalid-name
        self.ft = xml_parser.ft
        # Reason: "to" meets requirement of snake_case. pylint: disable=invalid-name
        self.to = xml_parser.to
        self.title = xml_parser.title
        self.station_id = xml_parser.station_id
        self.date = xml_parser.date
        self.area_id = xml_parser.area_id
        self.archive_status = ArchiveStatusId.ARCHIVABLE.value

    def mark_archivable(self) -> None:
        Program.mark(self, ArchiveStatusId.ARCHIVABLE)

    def mark_archiving(self) -> None:
        Program.mark(self, ArchiveStatusId.ARCHIVING)

    def mark_suspended(self) -> None:
        Program.mark(self, ArchiveStatusId.SUSPENDED)

    def mark_failed(self) -> None:
        Program.mark(self, ArchiveStatusId.FAILED)

    def mark_archived(self) -> None:
        Program.mark(self, ArchiveStatusId.ARCHIVED)

    @staticmethod
    def mark(program: Program, archive_id: ArchiveStatusId) -> None:
        with SessionManager() as session:
            selected_program = session.query(Program).with_for_update().filter_by(id=program.id).one()
            selected_program.archive_status = archive_id.value
            session.flush()

    @property
    def ft_string(self) -> str:
        if not self.ft:
            message = f"{self.ft=}"
            raise ValueError(message)
        return RadikoDatetime.encode(self.ft)

    @property
    def to_string(self) -> str:
        if not self.to:
            message = f"{self.to=}"
            raise ValueError(message)
        return RadikoDatetime.encode(self.to)

    @staticmethod
    def is_empty(target_date: datetime.date) -> bool:
        with SessionManager() as session:
            count = cast(tuple[int], session.query(func.count(Program.id)).filter(Program.date == target_date).one())
            return count == (0,)

    @staticmethod
    def find(keywords: list[str]) -> list[Program]:
        """This method select Store model from database."""
        with SessionManager() as session:
            list_condition_keyword = [Program.title.like(f"%{keyword}%") for keyword in keywords]
            return (
                session.query(Program)
                .filter(
                    and_(or_(*list_condition_keyword), Program.archive_status.is_(ArchiveStatusId.ARCHIVABLE.value)),
                )
                .order_by(Program.ft.asc())
                .all()
            )

    @staticmethod
    def delete(boundary_date: datetime.date) -> None:
        with SessionManager() as session:
            session.query(Program).filter(Program.date < boundary_date).delete()



"""Stream spec factory."""
from logging import getLogger
from pathlib import Path
from typing import TYPE_CHECKING

# noinspection PyPackageRequirements
import ffmpeg
from radikoplaylist import MasterPlaylistClient, TimeFreeMasterPlaylistRequest

if TYPE_CHECKING:
    from asyncffmpeg import StreamSpec

    from radikopodcast.database.models import Program


class RadikoStreamSpecFactory:
    """Stream spec factory."""

    def __init__(self, program: "Program") -> None:
        self.program = program
        if not self.program.area_id:
            message = f"{self.program.area_id=}"
            raise ValueError(message)
        self.area_id = self.program.area_id
        self.logger = getLogger(__name__)

    async def create(self) -> "StreamSpec":
        """Creates."""
        out_file_name = f"./output/{self.program.ft_string}_{self.program.station_id}_{self.program.title}.m4a"
        self.logger.debug("out file name: %s", out_file_name)
        if Path(out_file_name).exists():
            self.logger.error("File already exists. out_file_name = %s", out_file_name)
            message = f"File already exists. {out_file_name=}"
            raise FileExistsError(message)
        master_playlist_request = TimeFreeMasterPlaylistRequest(
            self.program.station_id,
            int(self.program.ft_string),
            int(self.program.to_string),
        )
        master_playlist = MasterPlaylistClient.get(master_playlist_request, area_id=self.area_id)
        stream = ffmpeg.input(master_playlist.media_playlist_url, headers=master_playlist.headers, copytb="1")
        return ffmpeg.output(stream, out_file_name, f="mp4", c="copy")



"""This module implements exceptions for this package."""


class Error(Exception):
    """Base class for exceptions in this module.

    @see https://docs.python.org/3/tutorial/errors.html#user-defined-exceptions
    """


class XmlParseError(Error):
    """Target XML is invalid."""



"""Download programs."""
from collections.abc import Generator
from datetime import date, datetime, timedelta
from logging import getLogger

from radikopodcast.database.models import Program
from radikopodcast.radiko_datetime import RadikoDatetime
from radikopodcast.radikoapi.radiko_api import RadikoApi
from radikopodcast.radikoxml.xml_converter import XmlConverterProgram


class ProgramDownloader:
    """Download programs."""

    AREA_ID_DEFAULT = "JP13"  # TOKYO

    def __init__(self, base_datetime: datetime, *, area_id: str = AREA_ID_DEFAULT) -> None:
        self.time_free_oldest_date = RadikoDatetime.time_free_oldest_date(base_datetime)
        self.time_free_day_before_newest_date = RadikoDatetime.time_free_day_before_newest_date(base_datetime)
        self.area_id = area_id
        self.logger = getLogger(__name__)

    def download_all_time_free_programs(self) -> None:
        """Download programs from radiko API."""
        for target_date in self.daterange(self.time_free_oldest_date, self.time_free_day_before_newest_date):
            self.logger.debug("target_date = %04d-%02d-%02d", target_date.year, target_date.month, target_date.day)
            if Program.is_empty(target_date):
                element_tree_radiko = RadikoApi(area_id=self.area_id).get_program(target_date)
                Program.save_all(XmlConverterProgram(target_date, element_tree_radiko, self.area_id).to_model())

    @staticmethod
    def daterange(start_date: date, end_date: date) -> Generator[date, None, None]:
        for day in range(int((end_date - start_date).days) + 1):
            yield start_date + timedelta(day)



"""Configuration."""
from dataclasses import dataclass, field

from yamldataclassconfig.config import YamlDataClassConfig


@dataclass
class Config(YamlDataClassConfig):
    """Configuration."""

    # Reason: To use auto complete
    area_id: str = "JP13"
    number_process: int = 3
    stop_if_file_exists: bool = False
    keywords: list[str] = field(default_factory=list)



"""Test for RadikoDateTime."""
from datetime import date, datetime

from freezegun.api import freeze_time
import pytest

from radikopodcast.radiko_datetime import JST, RadikoDate, RadikoDatetime


class TestRadikoDatetime:
    """Test for RadikoDatetime."""

    CASE_TIME_FREE_OLDEST_DAY = [
        (datetime(2021, 1, 7, 21, 55, 15, tzinfo=JST), date(2020, 12, 31)),
        (datetime(2021, 1, 8, 4, 59, 59, tzinfo=JST), date(2020, 12, 31)),
        (datetime(2021, 1, 8, 5, 0, 0, tzinfo=JST), date(2021, 1, 1)),
    ]
    CASE_TIME_FREE_DAY_BEFORE_NEWEST_DAY = [
        (datetime(2021, 1, 7, 21, 55, 15, tzinfo=JST), date(2021, 1, 6)),
        (datetime(2021, 1, 8, 4, 59, 59, tzinfo=JST), date(2021, 1, 6)),
        (datetime(2021, 1, 8, 5, 0, 0, tzinfo=JST), date(2021, 1, 7)),
    ]
    CASE_IS_SAME_RADIKO_DATE = [
        (datetime(2021, 11, 21, 4, 59, 59, tzinfo=JST), datetime(2021, 11, 20, 5, 0, 0, tzinfo=JST), True),
        (datetime(2021, 11, 21, 4, 59, 59, tzinfo=JST), datetime(2021, 11, 21, 5, 0, 0, tzinfo=JST), False),
        (datetime(2021, 11, 20, 5, 0, 0, tzinfo=JST), datetime(2021, 11, 21, 4, 59, 59, tzinfo=JST), True),
        (datetime(2021, 11, 21, 5, 0, 0, tzinfo=JST), datetime(2021, 11, 21, 4, 59, 59, tzinfo=JST), False),
    ]

    @staticmethod
    def test_encode() -> None:
        assert RadikoDatetime.encode(datetime(2023, 3, 12, 17, 59, 25, tzinfo=JST)) == "20230312175925"

    @staticmethod
    def test_decode() -> None:
        assert RadikoDatetime.decode("20230312175925") == datetime(2023, 3, 12, 17, 59, 25, tzinfo=JST)

    @staticmethod
    @freeze_time("2021-11-13 00:00:00", tz_offset=-9)
    def test_download_if_program_has_not_been_downloaded() -> None:
        now = RadikoDatetime.now_jst()
        assert now.strftime("%Y/%m/%d %H:%M:%S %Z") == "2021/11/13 00:00:00 JST"

    @staticmethod
    @pytest.mark.parametrize(
        ("argument", "expect"),
        CASE_TIME_FREE_OLDEST_DAY,
    )
    def test_time_free_oldest_day(argument: datetime, expect: date) -> None:
        assert RadikoDatetime.time_free_oldest_date(argument) == expect

    @staticmethod
    @pytest.mark.parametrize(
        ("argument", "expect"),
        CASE_TIME_FREE_DAY_BEFORE_NEWEST_DAY,
    )
    def test_time_free_day_before_newest_day(argument: datetime, expect: date) -> None:
        assert RadikoDatetime.time_free_day_before_newest_date(argument) == expect

    @staticmethod
    @pytest.mark.parametrize(
        ("datetime_a", "datetime_b", "expect"),
        CASE_IS_SAME_RADIKO_DATE,
    )
    def test_is_same_radiko_date(datetime_a: datetime, datetime_b: datetime, *, expect: bool) -> None:
        assert RadikoDatetime.is_same_radiko_date(datetime_a, datetime_b) == expect


class TestRadikoDate:
    """Test for RadikoDate."""

    @staticmethod
    def test_encode() -> None:
        assert RadikoDate.encode(date(2021, 11, 20)) == "20211120"

    @staticmethod
    def test_decode() -> None:
        assert RadikoDate.decode("20211120") == date(2021, 11, 20)
